function varargout=bench(Lfun,setfun,varargin)
% FUNCTION fc_bench.bench
% -- bench(LFUN,SETFUN)
%     Runs benchmark for each function given in the cell array LFUN. The 
%     function handle SETFUN is used to set input datas to these functions. 
%     There is the imposed syntax:
%        function [INPUTS,BDS]=SETFUN(N,verbose,varargin)
%          ...
%        end
%     By default, for all N in 5:5:20, computational time in second of 
%     each function in LFUN with INPUTS given by
%        [INPUTS,BDS]=SETFUN(N,verbose,varargin{:})  for all N in 5:5:20
%     is evaluated by tic-toc command:
%        t=tic(); OUT=LFUN{i}( INPUTS{:} ); tcpu=toc(t);
%     
%     Example to benchmark the function MTIMES (matricial product) between two 
%       N-by-N matrices for N in 50:50:200: 
%         LFUN={ @(X,Y)mtimes(X,Y) };  % or LFUN={ @mtimes };
%         SETFUN=@(varargin) set_matprod(varargin{:}); % or SETFUN=@set_matprod
%         fc_bench.bench(LFUN, SETFUN, 'LN', 50:50:200); 
%     
%       The 'set_matprod' function  as second parameter is given by
%         function [Inputs,bDs]=set_matprod(N,verbose,varargin)
%           X=randn(N,N); Y=randn(N,N);
%           bDs(1)=fc_bench.bdata('N',N,'%d',5);
%           Inputs={X,Y};
%         end

% -- bench(Lfun,setfun,key,value,...)
%
%     See User's Guide
%
  p = inputParser;
  p.KeepUnmatched=true; 
  %p.PartialMatching=false;
  p.addParamValue('LN',5:5:20);
  p.addParamValue('nbruns',5);
  p.addParamValue('debug',false,@islogical);
  p.addParamValue('error',[]); % Use only if all
  p.addParamValue('info',true,@islogical);
  p.addParamValue('labelsinfo',false,@islogical);
  p.addParamValue('caller',false,@islogical);
  p.addParamValue('save',false,@islogical);
  p.addParamValue('comment','',@(x) ischar(x) || iscell(x));
  p.addParamValue('savefile','',@ischar);
  p.addParamValue('savedir','',@ischar); % relative to currrent path. default 'benchs'
  p.addParamValue('threads',[]); % same length as Lfun : to specify number of threads
  p.addParamValue('names',{},@iscell); % same length as Lfun : to specify name of each function
  
  p.parse(varargin{:});
  R=p.Results;
  varargin=fc_tools.utils.deleteCellOptions(varargin,p.Parameters);
  Debug=R.debug;nbruns=R.nbruns;LN=R.LN;
  if isnumeric(LN)
    if size(LN,2)==1,LN=LN.';end
  end
  DateNum=now;
  if R.save || ~isempty(R.savefile) || ~isempty(R.savedir)
    savedir=R.savedir;
    if isempty(savedir), savedir='benchs';end % default directory
    savefile=R.savefile;
    if isempty(R.savefile)
      name=fc_bench.funname_small(Lfun{1});
      I=strfind(name,'_');
      if length(I)>0, name=name(1:I(end)-1);end
      Date=datestr(now,'yyyy-mm-dd_HH-MM-SS');
      [success,message,messageid]=mkdir('.',savedir);
      if ~success, error(message);end
      savefile=fullfile(savedir,['bench_',name,'_',Date,'.out']);
    else
      [pathstr,filename,ext] = fileparts(R.savefile);
      if isempty(pathstr)
        [success,message,messageid]=mkdir('.',savedir);
        if ~success, error(message);end
        savefile=fullfile(savedir,R.savefile);
      else
        [success,message,messageid]=mkdir('.',pathstr);
        if ~success, error(message);end
        savefile=R.savefile;
      end
    end
    R.save=true;
    [fid,res]=fopen(savefile,'w');
    assert(fid>=0,res)
    Fprintf=@(varargin) EchoingFprintf(fid,varargin{:});
  else
    Fprintf=@(varargin) fprintf(varargin{:});
  end
  nlabels=[];
  if ~isempty(R.threads)
    assert(length(R.threads)==length(Lfun) )
    if fc_tools.comp.isOctave() 
      setThreads=@(n) fc_bench.MKL_setNumThreads(n);
    else
      setThreads=@(n) maxNumCompThreads(n);
    end
  end
  nfun=length(Lfun);
  % set names
  names=R.names;
  if isempty(names), names=cell(1,nfun);end
  assert(length(names)==nfun)  
  for i=1:nfun
    if isempty(names{i})
      names{i}=fc_bench.funname_small(Lfun{i}); 
    end
  end
  
  ltd=fc_tools.utils.line_text_delimiter();
  format='';
  Fprintf(ltd);
  if R.info
    %Fprintf(ltd);
    fc_bench.print_info('fprintf',Fprintf);
    Fprintf(ltd);
  end
  % Print comment
  if length(R.comment)>0
    if ischar(R.comment)
      Fprintf(R.comment);
      if ~strcmp(R.comment(end-1:end),'\n'), Fprintf('\n');end
    else % iscell
      for i=1:length(R.comment)
        Fprintf(R.comment{i});
        if ~strcmp(R.comment{i}(end-1:end),'\n'), Fprintf('\n');end
      end
    end
    Fprintf(ltd);
  end
  if R.labelsinfo
    Fprintf('# Benchmarking functions:\n')
    for i=1:nfun
      Fprintf('#  fun[%d], %8s: %s\n',i-1,names{i},func2str(Lfun{i}))
    end
    if ~isempty(R.error)
      Fprintf('# cmpErr[i], error between fun[0] and fun[i] outputs computed with function\n#    %s\n# where\n',func2str(R.error))
      Fprintf('#    - 1st input parameter is the output of fun[0]\n')
      Fprintf('#    - 2nd input parameter is the output of fun[i]\n')
    end
    Fprintf(ltd);
  end
  
  if size(LN,1)==1
    [datas,bDs]=setfun(min(LN),true,'fprintf',Fprintf,varargin{:});
    if iscell(LN), getN=@(LN,i) LN{i};else;  getN=@(LN,i) LN(i);end
    %if iscell(LN), getN=@(i) LN{i};else;  getN=@(i) eval(LN(i));end
    nLN=size(LN,2);
  else
    [datas,bDs]=setfun(LN(1,:),true,'fprintf',Fprintf,varargin{:});
    if iscell(LN), getN=@(LN,i) {LN{i,:}};else;  getN=@(LN,i) LN(i,:);end
    %if iscell(LN), getN=@(i) {LN{i,:}};else;  getN=@(i) LN(i,:);end
    nLN=size(LN,1);
  end
  Res=cell(nLN+1,length(bDs));
  Fprintf(ltd);
  if R.save
    Fprintf('#benchfile: %s\n',savefile);
  end
  if R.caller
    [callerFileName,callerLineNumber]=fc_tools.utils.getCaller('level',3);
    if length(callerFileName)>0
      Fprintf('#caller: %s:%d\n',callerFileName,callerLineNumber);
    else
      str=string_command(Lfun,setfun,varargin{:});
      Fprintf('#caller: from command line, using fc_bench.bench function\n');
      Fprintf('#   -> %s\n',str);
    end
  end
  
  isTitle=false;
  for i=1:nLN
    %N=getN(i);
    N=getN(LN,i);
    [datas,bDs]=setfun(N,false,varargin{:});
    for j=1:length(bDs), Res{i+1,j}=bDs(j).value;end
    j=length(bDs)+1;
    if fc_tools.utils.isfunhandle(datas{end})
      funerror=datas{end};
      datas={datas{1:end-1}};
    else
      funerror=[];
    end
    fun=Lfun{1};
    if ~isempty(R.threads)
      name=sprintf('%0d_threads(s)',R.threads(1));
      if fc_tools.comp.isOctave()
        fc_bench.MKL_setNumThreads(R.threads(1));
      else
        maxNumCompThreads(R.threads(1));
      end
    else
      name=[names{1},'(s)']; % time in second
    end
    [tcpu,OutRef]=fc_bench.mean_run(fun,datas,nbruns);
    bDs(end+1)=fc_bench.bdata(name,tcpu,'%.3f',length(name)+3);
    Res{i+1,j}=bDs(j).value;j=j+1;
    if ~isempty(funerror)
      if iscell(OutRef)
        E=funerror(OutRef{:});
      else
        E=funerror(OutRef);
      end
      name=sprintf('  Error[0]');
      bDs(end+1)=fc_bench.bdata(name,E,'%.3e',length(name)+3);
      Res{i+1,j}=bDs(j).value;j=j+1;
    end
    for k=2:nfun
      fun=Lfun{k};
      if ~isempty(R.threads)
        if fc_tools.comp.isOctave()
          fc_bench.MKL_setNumThreads(R.threads(k));
        else
          maxNumCompThreads(R.threads(k));
        end
        name=sprintf('%0d_threads(s)',R.threads(k));
      else
        name=[names{k},'(s)']; % time in second
      end
      
      [tcpu,Out]=fc_bench.mean_run(fun,datas,nbruns);
      
    
      bDs(end+1)=fc_bench.bdata(name,tcpu,'%.3f',length(name)+3);
      Res{i+1,j}=bDs(j).value;j=j+1;
      if ~isempty(funerror)
        if iscell(Out), E=funerror(Out{:}); else, E=funerror(Out); end
        name=sprintf('  Error[%d]',k-1);
        bDs(end+1)=fc_bench.bdata(name,E,'%.3e',length(name)+3);
        Res{i+1,j}=bDs(j).value;j=j+1;
      end
      if ~isempty(R.error) % Comparative error
        cmpErr=R.error(OutRef,Out);
        name=sprintf('  cmpErr[%d]',k-1);
        bDs(end+1)=fc_bench.bdata(name,cmpErr,'%.3e',length(name)+3);
        Res{i+1,j}=bDs(j).value;j=j+1;
      end
    end
    if ~isTitle
      [format,nlabels]=fc_bench.print_title(bDs,DateNum,nbruns,'fprintf',Fprintf);
      isTitle=true;
    end
    fc_bench.print_bdatas(bDs,'fprintf',Fprintf,'nlabels',nlabels)
  end
  if R.save
    fclose(fid);
  end
  if nargout==1
    for j=1:size(Res,2), Res{1,j}=bDs(j).name;end
    varargout{1}=Res;
  end
end

function EchoingFprintf(fid,varargin)
  fprintf(varargin{:});
  fprintf(fid,varargin{:});
end

function str=string_command(Lfun,setfun,varargin)
  str='fc_bench.bench( { ';
  for i=1:length(Lfun)-1
    str=[str,fc_tools.utils.fun2str(Lfun{i}),', '];
  end
  str=[str,fc_tools.utils.fun2str(Lfun{end}),' },  '];
  str=[str,fc_tools.utils.fun2str(setfun),', ...'];
  str=[str,')'];
end
